import os
import logging
import timeit

import gmsh
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np
from scipy.interpolate import griddata
import matplotlib
import matplotlib.pyplot as plt

from fiqus.mains.MainPancake3D import Base
from fiqus.utils.Utils import GmshUtils
from fiqus.parsers.ParserGetDPTimeTable import ParserGetDPTimeTable
from fiqus.parsers.ParserGetDPOnSection import ParserGetDPOnSection

logger = logging.getLogger(__name__)


class TimeSeriesData:
    def __init__(
        self, timeValues, values, quantityName, units, title=None, fileName=None
    ):
        """ """
        self.t = timeValues
        self.values = values
        self.quantityName = quantityName
        self.units = units
        self.title = title
        self.fileName = fileName

    def plot(self, dir):
        plt.rcParams.update({"font.size": 10})
        plt.style.use("_mpl-gallery")
        fig, ax = plt.subplots(figsize=(5.95, 3))
        ax.plot(self.t, self.values, marker="o")
        ax.set_xlabel(f"Time (s)")
        ax.set_ylabel(f"{self.quantityName} ({self.units})")
        if self.title:
            ax.set_title(self.title)

        # Save the plot:
        if self.fileName is None:
            quantityName = self.quantityName.replace(" ", "_")
            filePath = os.path.join(dir, f"{quantityName}_TimeSeriesPlot.pdf")
        else:
            filePath = os.path.join(dir, f"{self.fileName}_TimeSeriesPlot.pdf")

        plt.savefig(filePath, bbox_inches="tight", pad_inches=0)
        plt.close("all")


class MagneticField:
    def __init__(
        self,
        pointsForMagnitudes,
        magnitudes,
        pointsForVectors,
        vectors,
        units,
        title=None,
        interpolationMethod="linear",
        colormap="jet",
        planeNormal=None,
    ):
        self.pointsForMagnitudes = pointsForMagnitudes
        self.magnitudes = magnitudes
        self.pointsForVectors = pointsForVectors
        self.vectors = vectors
        self.units = units
        self.title = title
        self.interpolationMethod = interpolationMethod
        self.colormap = colormap
        self.planeNormal = planeNormal

    def plot(self, dir, streamlines=True):
        start_time = timeit.default_timer()
        logger.info(f"Plotting {self.title} has been started.")
        plt.rcParams.update({"font.size": 10})
        plt.style.use("_mpl-gallery")

        points = self.pointsForMagnitudes
        magnitudes = self.magnitudes

        xArray = points[:, 0]
        yArray = points[:, 1]

        cropFactorVertical = 0.75
        cropFactorHorizontal = 0.66
        num_of_grid_points_in_y = 90

        # Put parsed values on a grid:
        aspectRatio = (xArray.max() - xArray.min()) / (yArray.max() - yArray.min())
        xLinSpace = np.linspace(
            xArray.min() * cropFactorHorizontal,
            xArray.max() * cropFactorHorizontal,
            round(num_of_grid_points_in_y * aspectRatio),
        )
        yLinSpace = np.linspace(
            yArray.min() * cropFactorVertical,
            yArray.max() * cropFactorVertical,
            num_of_grid_points_in_y,
        )
        grid_x, grid_y = np.meshgrid(xLinSpace, yLinSpace)
        grid_z = griddata(points, magnitudes, (grid_x, grid_y), method="linear")

        # Then interpolate the grid values:
        grid_points = np.array([grid_x.flatten(), grid_y.flatten()]).T
        grid_values = grid_z.flatten()
        xLinSpace = np.linspace(
            xArray.min() * cropFactorHorizontal,
            xArray.max() * cropFactorHorizontal,
            round(num_of_grid_points_in_y * aspectRatio) * 6,
        )
        yLinSpace = np.linspace(
            yArray.min() * cropFactorVertical,
            yArray.max() * cropFactorVertical,
            num_of_grid_points_in_y * 6,
        )
        grid_x, grid_y = np.meshgrid(xLinSpace, yLinSpace)
        grid_z = griddata(
            grid_points,
            grid_values,
            (grid_x, grid_y),
            method=self.interpolationMethod,
        )

        fig, ax = plt.subplots(figsize=(5.95, 4.165))

        # cs = ax.contourf(
        #     grid_x, grid_y, grid_z, cmap=matplotlib.colormaps[self.colormap], levels=np.linspace(0, 0.8, 16)
        # )
        cs = ax.imshow(
            grid_z,
            cmap=matplotlib.colormaps[self.colormap],
            extent=[
                xArray.min() * cropFactorHorizontal * 1000,
                xArray.max() * cropFactorHorizontal * 1000,
                yArray.min() * cropFactorVertical * 1000,
                yArray.max() * cropFactorVertical * 1000,
            ],
            origin="lower",
            aspect="auto",
        )
        ax.set_aspect("equal")
        ax.set_xlabel("x (mm)")
        ax.set_ylabel("z (mm)")

        fig.colorbar(
            cs,
            ax=ax,
            label=f"Magnitude of the Magnetic Field ({self.units})",
            location="bottom",
        )

        # # Create a Rectangle patch for each coil:
        # rect1 = patches.Rectangle((5.0e-3*1000, -4.042e-3/2*1000), 120.0e-6*60*1000, 4.042e-3*1000, linewidth=2, edgecolor='k', facecolor='none')
        # rect2 = patches.Rectangle((-5.0e-3*1000, -4.042e-3/2*1000), -120.0e-6*60*1000, 4.042e-3*1000, linewidth=2, edgecolor='k', facecolor='none')
        # # Add the patch to the Axes
        # ax.add_patch(rect1)
        # ax.add_patch(rect2)

        if streamlines:
            points = self.pointsForVectors
            vectors = self.vectors
            uValues = vectors[:, 0]
            vValues = vectors[:, 1]
            grid_u = griddata(
                points, uValues, (grid_x, grid_y), method=self.interpolationMethod
            )
            grid_v = griddata(
                points, vValues, (grid_x, grid_y), method=self.interpolationMethod
            )

            ax.streamplot(
                grid_x * 1000,
                grid_y * 1000,
                grid_u,
                grid_v,
                color="white",
                linewidth=0.8,
                density=1.0,
            )

        if self.title:
            ax.set_title(self.title)

        # Save the plot:
        time = self.title.split(" ")[-2]
        plotFileName = os.path.join(
            dir, f"Magnetic_Field_On_A_Cut_Plane_at_{time}_s.pdf"
        )
        plt.savefig(plotFileName, bbox_inches="tight", pad_inches=0)
        plt.close("all")
        end_time = timeit.default_timer()
        time_span = end_time - start_time
        logger.info(f"Plotting {self.title} has been finished in {time_span:.2f} s.")


class Postprocess(Base):
    def __init__(
        self,
        fdm,
        geom_folder,
        mesh_folder,
        solution_folder,
    ) -> None:
        super().__init__(fdm, geom_folder, mesh_folder, solution_folder)

    def plotTimeSeriesPlots(self):
        start_time = timeit.default_timer()
        logger.info("Plotting time series plots has been started.")

        for timeSeriesPlot in self.pp.timeSeriesPlots:
            filePath = os.path.join(
                self.solution_folder, timeSeriesPlot.fileName + "-TimeTableFormat.txt"
            )
            if timeSeriesPlot.position.turnNumber is not None:
                title = (
                    f"{timeSeriesPlot.quantityProperName} at turn"
                    f" {timeSeriesPlot.position.turnNumber}"
                )
            elif timeSeriesPlot.position.x is not None:
                title = (
                    f"{timeSeriesPlot.quantityProperName} at"
                    f" ({timeSeriesPlot.position.x}, {timeSeriesPlot.position.y},"
                    f" {timeSeriesPlot.position.z})"
                )
            else:
                title = f"{timeSeriesPlot.quantityProperName}"

            # Parse data:
            parser = ParserGetDPTimeTable(filePath)
            timeValues = parser.time_values
            scalarvalues = parser.get_equivalent_scalar_values()

            if parser.data_type == "vector":
                title = title + " (Magnitudes of 3D Vectors)"

            elif parser.data_type == "tensor":
                title = title + " (Von Misses Equivalents of Rank 2 Tensors)"

            data = TimeSeriesData(
                timeValues=timeValues,
                values=scalarvalues,
                quantityName=timeSeriesPlot.quantityProperName,
                units=timeSeriesPlot.units,
                title=title,
            )
            data.plot(dir=self.solution_folder)

        end_time = timeit.default_timer()
        time_span = end_time - start_time
        logger.info(
            f"Plotting time series plots has been finished in {time_span:.2f} s."
        )

        if self.python_postprocess_gui:
            gmsh.initialize()
            gmsh.option.setNumber("Geometry.Volumes", 0)
            gmsh.option.setNumber("Geometry.Surfaces", 0)
            gmsh.option.setNumber("Geometry.Curves", 0)
            gmsh.option.setNumber("Geometry.Points", 0)
            posFiles = [
                fileName
                for fileName in os.listdir(self.solution_folder)
                if fileName.endswith(".pos")
            ]
            for posFile in posFiles:
                gmsh.open(os.path.join(self.solution_folder, posFile))

            self.gu = GmshUtils(self.solution_folder)
            self.gu.launch_interactive_GUI()
            gmsh.finalize()

    def plotMagneticFieldOnCutPlane(self):
        start_time = timeit.default_timer()
        logger.info("Parsing magnetic field on the cut plane has been started.")

        magnitudeFilePath = os.path.join(
            self.solution_folder, "magneticFieldOnCutPlaneMagnitude-DefaultFormat.pos"
        )
        vectorFilePath = os.path.join(
            self.solution_folder, "magneticFieldOnCutPlaneVector-DefaultFormat.pos"
        )

        magnitudeParser = ParserGetDPOnSection(
            magnitudeFilePath, data_type="scalar", depth=1
        )
        magnitudeParser.project_values_on_a_plane(
            plane_normal=self.pp.magneticFieldOnCutPlane.planeNormal,
            plane_x_axis_unit_vector=self.pp.magneticFieldOnCutPlane.planeXAxisUnitVector,
        )

        if self.pp.magneticFieldOnCutPlane.streamLines:
            vectorParser = ParserGetDPOnSection(
                vectorFilePath, data_type="vector", depth=1
            )
            vectorParser.project_values_on_a_plane(
                plane_normal=self.pp.magneticFieldOnCutPlane.planeNormal,
                plane_x_axis_unit_vector=self.pp.magneticFieldOnCutPlane.planeXAxisUnitVector,
            )

        end_time = timeit.default_timer()
        time_span = end_time - start_time
        logger.info(
            "Parsing magnetic field on the cut plane has been finished in"
            f" {time_span:.2f} s."
        )
        for timeStep in range(2, len(magnitudeParser.time_values)):
            time = magnitudeParser.time_values[timeStep]
            if (
                self.pp.magneticFieldOnCutPlane.timesToBePlotted is None
                or time in self.pp.magneticFieldOnCutPlane.timesToBePlotted
            ):
                magnitudePoints = magnitudeParser.points
                magnitudes = magnitudeParser.get_values_at_time_step(timeStep)
                vectorPoints = vectorParser.points
                vectors = vectorParser.get_values_at_time_step(timeStep)

                magneticFieldOnCutPlane = MagneticField(
                    pointsForMagnitudes=magnitudePoints,
                    magnitudes=magnitudes,
                    pointsForVectors=vectorPoints,
                    vectors=vectors,
                    units="T",
                    title=(
                        "Magnetic Field On Plane"
                        f" {self.pp.magneticFieldOnCutPlane.planeNormal[0]}x + "
                        f" {self.pp.magneticFieldOnCutPlane.planeNormal[1]}y + "
                        f" {self.pp.magneticFieldOnCutPlane.planeNormal[2]}z = 0 at"
                        f" {time:.2f} s"
                    ),
                    interpolationMethod=self.pp.magneticFieldOnCutPlane.interpolationMethod,
                    colormap=self.pp.magneticFieldOnCutPlane.colormap,
                    planeNormal=self.pp.magneticFieldOnCutPlane.planeNormal,
                )
                magneticFieldOnCutPlane.plot(
                    dir=self.solution_folder,
                    streamlines=self.pp.magneticFieldOnCutPlane.streamLines,
                )



import gmsh
import statistics

class ParserMSH:
    def __init__(self, mesh_file_path):
        """
        Read msh file and returns mesh format and physical names as class attributes.
        :param mesh_file_path: Full path to .msh file, including file name and extension.
        """
        self.mesh_file_path = mesh_file_path

        self._mesh_format_markers = {'s': '$MeshFormat', 'e': '$EndMeshFormat'}
        self._physical_name_markers = {'s': 'PhysicalNames', 'e': '$EndPhysicalNames'}

        with open(mesh_file_path) as f:
            self._contents = f.read()

    def __get_content(self, markers_dict):
        """
        Gets text string between two markers specified in markers_dict
        """
        return self._contents[self._contents.find(markers_dict['s']) + len(markers_dict['s']):self._contents.find(markers_dict['e'])]

    def get_average_mesh_quality(self):
        """
        Gets the lowest mesh quality from the mesh file
        """
        gmsh.initialize()
        gmsh.open(self.mesh_file_path)

        # SICN not implemented in 1D!
        allElementsDim2 = gmsh.model.mesh.getElements(dim=2)[1]
        allElementsDim3 = gmsh.model.mesh.getElements(dim=3)[1]
        allElements = list(allElementsDim2[0]) + (list(allElementsDim3[0]) if allElementsDim3 else [])
        lowestQuality = statistics.fmean(gmsh.model.mesh.getElementQualities(allElements))

        gmsh.finalize()

        return lowestQuality

    @property
    def mesh_format(self):
        """
        Parse mesh_generators field and assign it to the class attribute
        """
        return self.__get_content(self._mesh_format_markers)

    @property
    def physical_names(self):
        """
        Parse physical_names field and assign it to the class attribute
        """
        return self.__get_content(self._physical_name_markers)



import os
import sys
import time

from fiqus.data.DataFiQuSPancake3D import (
    Pancake3DGeometry,
    Pancake3DMesh,
    Pancake3DSolve,
    Pancake3DPostprocess,
)

if len(sys.argv) == 3:
    sys.path.insert(0, os.path.join(os.getcwd(), "steam-fiqus-dev"))


class Base:
    """
    Base class for geometry, mesh, and solution classes. It is created to avoid code
    duplication and to make the code more readable. Moreover, it guarantees that
    all the classes have the same fundamental methods and attributes.
    """

    def __init__(
        self,
        fdm,
        geom_folder=None,
        mesh_folder=None,
        solution_folder=None,
    ) -> None:
        """
        Define the fundamental attributes of the class.

        :param fdm: fiqus data model
        :param geom_folder: folder where the geometry related files are stored
        :type geom_folder: str
        :param mesh_folder: folder where the mesh related files are stored
        :type mesh_folder: str
        :param solution_folder: folder where the solution related files are stored
        :type solution_folder: str
        """

        self.magnet_name = fdm.general.magnet_name
        self.geom_folder = geom_folder
        self.mesh_folder = mesh_folder
        self.solution_folder = solution_folder

        self.dm = fdm  # Data model
        self.geo: Pancake3DGeometry = fdm.magnet.geometry  # Geometry data
        self.mesh: Pancake3DMesh = fdm.magnet.mesh  # Mesh data
        self.solve: Pancake3DSolve = fdm.magnet.solve  # Solve data
        self.pp: Pancake3DPostprocess = fdm.magnet.postproc  # Postprocess data

        self.geo_gui = False
        self.mesh_gui = False
        self.solve_gui = False
        self.python_postprocess_gui = False

        if fdm.run.launch_gui:
            if fdm.run.type == "start_from_yaml":
                self.python_postprocess_gui = True
            elif fdm.run.type == "geometry_only":
                self.geo_gui = True
            elif fdm.run.type == "mesh_only":
                self.mesh_gui = True
            elif fdm.run.type == "geometry_and_mesh":
                self.mesh_gui = True
            elif fdm.run.type == "solve_only":
                self.solve_gui = True
            elif fdm.run.type == "post_process_getdp_only":
                self.solve_gui = True
            elif fdm.run.type == "solve_with_post_process_python":
                self.python_postprocess_gui = True
            elif fdm.run.type == "post_process_python_only":
                self.python_postprocess_gui = True

        # Geometry related files:
        if self.geom_folder is not None:
            self.brep_file = os.path.join(self.geom_folder, f"{self.magnet_name}.brep")
            self.vi_file = os.path.join(self.geom_folder, f"{self.magnet_name}.vi")
            self.geometry_data_file = os.path.join(self.geom_folder, "geometry.yaml")

        # Mesh related files:
        if self.mesh_folder is not None:
            self.mesh_file = os.path.join(self.mesh_folder, f"{self.magnet_name}.msh")
            self.regions_file = os.path.join(
                self.mesh_folder, f"{self.magnet_name}.regions"
            )
            self.mesh_data_file = os.path.join(self.mesh_folder, "mesh.yaml")
        # Solution related files:
        if self.solution_folder is not None:
            self.pro_file = os.path.join(
                self.solution_folder, f"{self.magnet_name}.pro"
            )


from fiqus.geom_generators.GeometryPancake3D import Geometry
from fiqus.mesh_generators.MeshPancake3D import Mesh
from fiqus.getdp_runners.RunGetdpPancake3D import Solve
from fiqus.post_processors.PostProcessPancake3D import Postprocess


class MainPancake3D:
    """
    The main class for working with simulations for high-temperature superconductor
    pancake coil magnets.

    Geometry can be created and saved as a BREP file. Parameters like the number of
    turns, tape dimensions, contact layer thicknesses, and other dimensions can be
    specified. Contact layers can be modeled as two-dimensional shells or three-dimensional
    volumes. Moreover, multiple pancakes can be stacked on top of each other.

    Using the BREP file created, a mesh can be generated and saved as an MSH file.
    Winding mesh can be structured, and parameters like, azimuthal number of elements
    per turn, axial number of elements, and radial number of elements per turn can be
    specified for each pancake coil. The appropriate regions will be assigned to the
    relevant volumes accordingly so that finite element simulations can be done.

    Using the mesh files, GetDP can be used to analyze Pancake3D coils.

    :param fdm: FiQuS data model
    """

    def __init__(self, fdm, verbose):
        self.fdm = fdm
        self.GetDP_path = None

        self.geom_folder = None
        self.mesh_folder = None
        self.solution_folder = None

    def generate_geometry(self, gui=False):
        """
        Generates the geometry of the magnet and save it as a BREP file. Moreover, a
        text file with the extension VI (volume information file) is generated, which
        stores the names of the volume tags in JSON format.
        """
        geometry = Geometry(
            self.fdm,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )

        geometry.generate_geometry()
        geometry.generate_vi_file()

        self.model_file = geometry.brep_file

    def load_geometry(self, gui=False):
        """
        Loads the previously generated geometry from the BREP file.
        """
        geometry = Geometry(
            self.fdm,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )

        geometry.load_geometry()
        self.model_file = geometry.brep_file

    def pre_process(self, gui=False):
        pass

    def mesh(self, gui=False):
        """
        Generates the mesh of the magnet, creates the physical regions, and saves it as
        an MSH file. Moreover, a text file with the extension REGIONS is generated,
        which stores the names and tags of the physical regions in YAML format.
        """
        mesh = Mesh(
            self.fdm,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )

        mesh.generate_mesh()
        mesh.generate_regions()
        mesh.generate_mesh_file()

        self.model_file = mesh.mesh_file

        return {"gamma": 0}  # to be modified with mesh_parameters (see multipole)

    def load_mesh(self, gui=False):
        """
        Loads the previously generated mesh from the MSH file.
        """
        mesh = Mesh(
            self.fdm,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )
        mesh.load_mesh()

        self.model_file = mesh.mesh_file

    def solve_and_postprocess_getdp(self, gui=False):
        """
        Simulates the Pancake3D magnet with GetDP using the created mesh file and post
        processes the results.
        """
        solve = Solve(
            self.fdm,
            self.GetDP_path,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )
        solve.assemble_pro()

        start_time = time.time()
        solve.run_getdp(solve=True, postOperation=True)

        return time.time() - start_time

    def post_process_getdp(self, gui=False):
        solve = Solve(
            self.fdm,
            self.GetDP_path,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )
        solve.assemble_pro()

        start_time = time.time()
        solve.run_getdp(solve=False, postOperation=True)

        return time.time() - start_time

    def post_process_python(self, gui=False):
        """
        To be written.
        """
        postprocess = Postprocess(
            self.fdm,
            self.geom_folder,
            self.mesh_folder,
            self.solution_folder,
        )

        if self.fdm.magnet.postproc is not None:
            if self.fdm.magnet.postproc.timeSeriesPlots is not None:
                postprocess.plotTimeSeriesPlots()
            if self.fdm.magnet.postproc.magneticFieldOnCutPlane is not None:
                postprocess.plotMagneticFieldOnCutPlane()

        return {"overall_error": 0}

    def plot_python(self):
        pass
